from typing import Dict
from typing import List

from cloudspec import CloudSpecFunction
from cloudspec import CloudSpecParam


def return_type(hub, func_data: CloudSpecFunction):
    ret = ""
    if func_data.return_type:
        ret += "\n\n    Returns:\n"
        ret += f"        {func_data.return_type}"
        ret += "\n"
    return ret


def doc(hub, func_data: CloudSpecFunction):
    if func_data.doc:
        doc = func_data.doc.replace('"""', "'''")
        doc = "\n" + hub.tool.format.wrap.indent(doc, 1) + "\n"
    else:
        doc = ""
    return doc


def parse(hub, functions: Dict[str, CloudSpecFunction], targets: List[str]):
    funcs = {}
    if not functions:
        # It might be an init file
        return funcs
    for function_name in targets:
        function_data = functions.get(function_name)
        # This lets us create partial plugin
        if function_data:
            # Add resource_id parameter
            if (
                function_name
                in ["present", "absent", "get", "create", "update", "delete"]
                and "resource_id"
                not in hub.cloudspec.parse.param.simple_map(function_data.params).keys()
            ):
                function_data.params["resource_id"] = CloudSpecParam(
                    hub=hub,
                    name="resource_id",
                    doc=f"{function_data.hardcoded.get('resource_name', '').capitalize()} unique ID",
                    required=True
                    if function_name in ["update", "delete", "get"]
                    else False,
                    target="kwargs",
                    target_type="mapping",
                    param_type="str",
                    exclude_from_example=True,
                )
            # Add name parameter
            if (
                function_name
                in ["present", "absent", "get", "list", "create", "update", "delete"]
                and "name"
                not in hub.cloudspec.parse.param.simple_map(function_data.params).keys()
            ):
                function_data.params["name"] = CloudSpecParam(
                    hub=hub,
                    name="name",
                    doc=f"Idem name of the resource.",
                    required=True if function_name in ["present", "absent"] else False,
                    target="kwargs",
                    target_type="mapping",
                    param_type="str",
                    exclude_from_example=True,
                )

            doc = hub.cloudspec.parse.function.doc(function_data)
            doc += hub.cloudspec.parse.param.sphinx_docs(function_data.params)
            doc += "\n\n    Returns:\n        Dict[str, Any]\n"

            if "absent" == function_name and function_data.params:
                # By idem's contract,
                #   - all absent function parameters except 'name' should not be required
                for param_name, metadata in function_data.params.items():
                    if param_name.lower() != "name":
                        metadata["required"] = False

            param_mapping = hub.cloudspec.parse.param.mappings(function_data.params)

            funcs[function_name] = dict(
                function=dict(
                    name=function_name,
                    hardcoded=function_data.hardcoded,
                    doc=doc,
                    header_params=hub.cloudspec.parse.param.headers(
                        function_data.params
                    ),
                    required_call_params=hub.cloudspec.parse.param.required_call_params(
                        function_data.params
                    ),
                ),
                parameter=dict(
                    mapping=param_mapping,
                ),
            )
    return funcs



"""State module for a sample state.

.. todo:: (Set ``todo_include_todos = False`` in ``docs/conf.py`` to hide rst todos)
    A sample state using the sample idem_cloud modules for an idea of what a state module should look like.
    All states should have a "present", "absent" and "describe" functions, nothing else.
    Any additional functionality that needs to be exposed via Idem CLI, should be added to Exec module.
"""
from typing import Dict, Any
import dict_tools.differ as differ

__contracts__ = ["resource"]  # , "soft_fail"]

async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    description: str = None
) -> Dict[str, Any]:
    """Ensure that a resource exists and is in the correct state.

    Args:
        name(str):
            The state name.

        resource_id(str):
            An identifier of the resource in the provider. Defaults to None.

        description(str, Optional):
            Description of the resource in the provider. Defaults to None.

    Returns:
        Dict[str, Any]

    Example:
        .. code-block:: sls

          my_state_name:
            {{cookiecutter.service_name}}.sample.present:
              - name: a_sample_name
              - description: Managed by Idem
    """
    result = dict(
        comment=[], old_state={}, new_state={}, name=name, result=True, rerun_data=None
    )

    desired_state = {
        k: v
        for k, v in locals().items()
        if k not in ("hub", "ctx", "kwargs", "result") and v is not None
    }

    if resource_id:
        before = await hub.exec.{{cookiecutter.service_name}}.sample.get(
            ctx,
            name=name,
            resource_id=resource_id,
        )

        if not before["result"] or not before["ret"]:
            result["result"] = False
            result["comment"] = before["comment"]
            return result

        result["old_state"] = before.ret

        result["comment"].append(
            f"{{cookiecutter.service_name}}.sample: {name} already exists"
        )

        # If there are changes in desired state from existing state
        changes = differ.deep_diff(before.ret if before.ret else {}, desired_state)

        if bool(changes.get("new")):
            if ctx.test:
                # If `idem state` was run with the `--test` flag, then don't actually make any changes in the provider
                result["new_state"] = hub.tool.{{cookiecutter.service_name}}.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state=desired_state
                )
                result["comment"] = (
                    f"Would update {{cookiecutter.service_name}}.sample: {name}"
                )
                return result
            else:
                # Update the resource
                update_ret = await hub.exec.{{cookiecutter.service_name}}.sample.update(
                    ctx,
                    name=name,
                    resource_id=resource_id,
                    description=description,
                )
                result["result"] = update_ret["result"]

                if result["result"]:
                    result["comment"].append(
                        f"Updated '{{cookiecutter.service_name}}.sample: {name}'")
                else:
                    result["comment"].append(update_ret["comment"])
    else:
        if ctx.test:
            result["new_state"] = hub.tool.hub.tool.{{cookiecutter.service_name}}.generate_test_state(
                enforced_state={},
                desired_state=desired_state
            )
            result["comment"] = (
            f"Would create {{cookiecutter.service_name}}.sample: {name}",)
            return result
        else:
            create_ret = await hub.exec.{{cookiecutter.service_name}}.sample.create(
                ctx,
                name=name,
                description=description
            )
            result["result"] = create_ret["result"]

            if result["result"]:
                result["comment"].append(
                    f"Created '{{cookiecutter.service_name}}.sample: {name}'")
                resource_id = create_ret["ret"]["resource_id"]
                # Safeguard for any future errors so that the resource_id is saved in the ESM
                result["new_state"] = dict(name=name, resource_id=resource_id)
            else:
                result["comment"].append(create_ret["comment"])

    if not result["result"]:
        # If there is any failure in create/update, it should reconcile.
        # The type of data is less important here to use default reconciliation
        # If there are no changes for 3 runs with rerun_data, then it will come out of execution
        result["rerun_data"] = dict(name=name, resource_id=resource_id)

    after = await hub.exec.{{cookiecutter.service_name}}.sample.get(
        ctx,
        name=name,
        resource_id=resource_id,
    )
    result["new_state"] = after.ret
    return result


async def absent(hub, ctx, name: str, resource_id: str = None):
    """Delete resource in the cloud provider

    Args:
        name(str):
            The state name.

        resource_id(str):
            An identifier of the resource in the provider. Defaults to None.

    Returns:
        Dict[str, Any]

    Example:
        .. code-block:: sls

          my_state_name:
            {{cookiecutter.service_name}}.sample.absent:
              - name: item_name
    """
    result = dict(
        comment=[], old_state={}, new_state={}, name=name, result=True, rerun_data=None
    )

    if not resource_id:
        result["comment"].append(
            f"'{{cookiecutter.service_name}}.sample: {name}' already absent")
        return result

    before = await hub.exec.{{cookiecutter.service_name}}.sample.get(
        ctx,
        name=name,
        resource_id=resource_id,
    )

    if before["ret"]:
        if ctx.test:
            result["comment"] = f"{{cookiecutter.service_name}}.sample: {name}"
            return result

        delete_ret = await hub.exec.{{cookiecutter.service_name}}.sample.delete(
            ctx,
            name=name,
            resource_id=resource_id,
        )
        result["result"] = delete_ret["result"]

        if result["result"]:
            result["comment"].append(
                f"Deleted {{cookiecutter.service_name}}.sample: {name}")
        else:
            # If there is any failure in delete, it should reconcile.
            # The type of data is less important here to use default reconciliation
            # If there are no changes for 3 runs with rerun_data, then it will come out of execution
            result["rerun_data"] = resource_id
            result["comment"].append(delete_ret["result"])
    else:
        result["comment"].append(
            f"{{cookiecutter.service_name}}.sample: {name} already absent")
        return result

    result["old_state"] = before.ret
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describes one or more sample states in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe {{cookiecutter.service_name}}.sample
    """
    result = {}

    ret = await hub.exec.{{cookiecutter.service_name}}.sample.list(
        ctx
    )

    if not ret or not ret["result"]:
        hub.log.debug(f"Could not describe {{cookiecutter.service_name}}.sample {ret['comment']}")
        return result

    for resource in ret["ret"]:
        resource_id = resource.get("resource_id")
        result[resource_id] = {
            "{{cookiecutter.service_name}}.sample.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



"""Exec module for managing sample resources."""

# This will put our "list_" function on the hub as "list"
# Func alias works around function names that shadow builtin python names
__func_alias__ = {"list_": "list"}


async def get(hub, ctx, name: str, resource_id: str):
    """Get a sample resource via resource_id.

    Args:
        name(str):
            The name of the Idem state.

        resource_id(str):
            An identifier of the resource in the provider.

    Returns:
        Dict[str, Any]:
            Returns sample resource in present format

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec {{cookiecutter.service_name}}.sample.get name=my_resource resource_id=resource_1

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: {{cookiecutter.service_name}}.sample.get
                - kwargs:
                    name: my_resource
                    resource_id: resource_1
    """
    return dict(
        comment=["sample get"],
        ret={
            "name": name,
            "resource_id": resource_id,
            "description": "Sample description"
        },
        result=True)


async def list_(hub, ctx, name: str = None):
    """List sample resources.

    Args:
        name(str, Optional):
            The name of the Idem state.

    Returns:
        Dict[str, Any]:
            Returns sample resources in present format

    Examples:
        Calling this exec module function from the cli with filters

        .. code-block:: bash

            idem exec {{cookiecutter.service_name}}.sample.list

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resources:
              exec.run:
                - path: {{cookiecutter.service_name}}.sample.list
    """
    return dict(
        comment=["sample get"],
        ret=[{
            "name": name,
            "resource_id": "resource_1",
            "description": "Sample description"
        }],
        result=True)


async def create(hub, ctx, name: str, resource_id: str = None, description: str = None):
    """Create a sample resource.

    Args:
        name(str):
            The state name.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        description(str, Optional):
            Description of the resource in the provider. Defaults to None.

    Returns:
        Dict[str, Any]:
            Returns sample resource in present format

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec {{cookiecutter.service_name}}.sample.create name=my_resource description="Managed by Idem"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: {{cookiecutter.service_name}}.sample.create
                - kwargs:
                    name: my_resource
                    description: Managed by Idem
    """
    return dict(
        comment=["sample create"],
        ret={
                "name": name,
                "resource_id": resource_id if resource_id else "resource_1",
                "description": description if description else "Sample description",
            },
        result=True)


async def update(hub, ctx, name: str, resource_id: str, description: str = None):
    """Update a sample resource.

    Args:
        name(str):
            The state name.

        resource_id(str):
            An identifier of the resource in the provider.

        description(str, Optional):
            Description of the resource in the provider. Defaults to None.

    Returns:
        Dict[str, Any]:
            Returns sample resource in present format

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec {{cookiecutter.service_name}}.sample.update name=my_resource resource_id=resource_1 description="Managed by Idem"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: {{cookiecutter.service_name}}.sample.update
                - kwargs:
                    resource_id: resource_1
                    name: my_resource
                    description: Managed by Idem
    """
    return dict(
        comment=["sample update"],
        ret={
                "name": name,
                "resource_id": resource_id,
                "description": description if description else "Sample description"
            },
        result=True)


async def delete(hub, ctx, name: str, resource_id: str):
    """Delete a sample resource.

    Args:
        name(str):
            The state name.

        resource_id(str):
            An identifier of the resource in the provider.

    Returns:
        Dict[str, Any]:
            Returns success or failure

    Examples:
        Calling this exec module function from the cli

        .. code-block:: bash

            idem exec {{cookiecutter.service_name}}.sample.delete name=my_resource resource_id=resource_1

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: {{cookiecutter.service_name}}.sample.delete
                - kwargs:
                    resource_id: resource_1
                    name: my_resource
    """
    return dict(
        comment=["sample delete"],
        ret={},
        result=True)
